import pathlib

import ase.io
import networkx as nx
import numpy as np
import znh5md
from ase.data.colors import jmol_colors
from ase.neighborlist import build_neighbor_list


def read_file(filename: str) -> ase.Atoms:
    """Reads a file and returns a networkx graph."""
    if pathlib.Path(filename).suffix == ".h5":
        return znh5md.ASEH5MD(filename).get_atoms_list()
    return ase.io.read(filename)


def _rgb2hex(data):
    r, g, b = np.array(data * 255, dtype=int)
    return "#%02x%02x%02x" % (r, g, b)


def get_graph(atoms: ase.Atoms) -> nx.Graph:
    """Returns a networkx graph from an ase.Atoms object."""
    atoms.pbc = False
    nl = build_neighbor_list(atoms, self_interaction=False)
    cm = nl.get_connectivity_matrix(sparse=False)
    G = nx.from_numpy_array(cm)

    node_data = {}

    for node in G.nodes:
        node_data[node] = {
            "symbol": atoms[node].symbol,
            "number": atoms[node].number.item(),
            "color": _rgb2hex(jmol_colors[atoms[node].number]),
            "radius": 0.25 * (2 - np.exp(-0.2 * atoms[node].number)),
            "position": atoms[node].position.tolist(),
        }

    nx.set_node_attributes(G, node_data)

    return G



"""Tools used for selecting objects."""
import ase
import networkx as nx

from zndraw import io


def select_identical_species(atoms: ase.Atoms, selected_ids: list[int]) -> list[int]:
    """Select all atoms with the same species as the selected atoms."""
    selected_ids = set(selected_ids)
    for idx in tuple(selected_ids):
        selected_symbol = atoms[idx].symbol
        selected_ids.update(
            idx for idx, atom in enumerate(atoms) if atom.symbol == selected_symbol
        )
    return list(selected_ids)


def select_connected(atoms: ase.Atoms, selected_ids: list[int]) -> list[int]:
    """Select all atoms connected to the selected atoms."""
    graph = io.get_graph(atoms)

    total_ids = []

    for node_id in selected_ids:
        total_ids += list(nx.node_connected_component(graph, node_id))

    return list(set(total_ids))



import uuid

import networkx as nx
import numpy as np
from flask import Flask, render_template, request, session

from zndraw import globals, io, tools

app = Flask(__name__)
app.secret_key = str(uuid.uuid4())


@app.route("/")
def index():
    """Render the main ZnDraw page."""
    session["key"] = str(uuid.uuid4())  # TODO use session key e.g. for atoms cache
    return render_template("index.html", config=globals.config.dict())


@app.route("/config", methods=["POST", "GET"])
def config():
    """Get the zndraw configuration."""
    if request.method == "POST":
        print(f"Updating config: {request.json}")
        for key, value in request.json.items():
            setattr(globals.config, key, value)
    return {
        **globals.config.dict(),
        "total_frames": len(globals.config._atoms_cache) - 1,
    }


@app.route("/graph", methods=["POST"])
def get_graph():
    step = request.json
    try:
        atoms = globals.config.get_atoms(step=int(step))
        graph = io.get_graph(atoms)
        return {
            "nodes": [{**graph.nodes[idx], "id": idx} for idx in graph.nodes],
            "edges": list(graph.edges),
            "box": atoms.get_cell().diagonal().tolist(),
        }
    except KeyError:
        return {}


@app.route("/data", methods=["POST"])
def positions_step():
    return tools.data.serialize_atoms(request.json["start"], request.json["stop"])


@app.route("/select", methods=["POST"])
def select() -> list[int]:
    """Update the selected atoms."""
    step = request.json["step"]
    method = request.json["method"]
    try:
        selected_ids = [int(x) for x in request.json["selected_ids"]]
    except TypeError:
        selected_ids = []
    if method in ["particles", "none"]:
        return {"selected_ids": selected_ids, "updated": False}

    atoms = globals.config.get_atoms(step)

    if method == "species":
        return {
            "selected_ids": tools.select.select_identical_species(atoms, selected_ids),
            "updated": True,
        }
    elif method == "connected":
        return {
            "selected_ids": tools.select.select_connected(atoms, selected_ids),
            "updated": True,
        }
    else:
        raise ValueError(f"Unknown selection method: {method}")


@app.route("/add_update_function", methods=["POST"])
def add_update_function():
    """Add a function to the config."""
    try:
        signature = globals.config.get_modifier_schema(request.json)
    except (ImportError, ValueError) as err:
        return {"error": str(err)}
    return signature


@app.route("/update", methods=["POST"])
def update_scene():
    """Update the scene with the selected atoms."""
    modifier = request.json["modifier"]
    modifier_kwargs = request.json["modifier_kwargs"]
    selected_ids = list(sorted(request.json["selected_ids"]))
    step = request.json["step"]
    points = np.array([[x["x"], x["y"], x["z"]] for x in request.json["points"]])
    globals.config.run_modifier(
        modifier, selected_ids, step, modifier_kwargs, points=points
    )
    return {}


@app.route("/add_analysis", methods=["POST"])
def add_analysis():
    """Add a function to the config."""
    import importlib

    try:
        module_name, function_name = request.json.rsplit(".", 1)
        module = importlib.import_module(module_name)
        cls = getattr(module, function_name)()
        schema = cls.schema_from_atoms(globals.config._atoms_cache)
        schema["title"] = request.json
    except (ImportError, ValueError) as err:
        return {"error": str(err)}
    return schema


@app.route("/analyse", methods=["POST"])
def analyse():
    import importlib

    selected_ids = list(sorted(request.json["selected_ids"]))
    step = request.json["step"]
    modifier = request.json["modifier"]
    modifier_kwargs = request.json["modifier_kwargs"]
    module_name, function_name = modifier.rsplit(".", 1)
    module = importlib.import_module(module_name)
    cls = getattr(module, function_name)
    instance = cls(**modifier_kwargs)
    fig = instance.run(selected_ids)
    return fig.to_json()


@app.route("/load")
def load():
    """Function to call asynchronously to load atoms in the background."""
    globals.config.load_atoms()
    return {}



import itertools

import ase
import numpy as np
import pandas as pd
import plotly.express as px
from pydantic import BaseModel

from zndraw import globals


class Distance(BaseModel):
    smooth: bool = False

    @classmethod
    def schema_from_atoms(cls, atoms):
        return cls.schema()

    def run(self, ids):
        atoms_lst = list(globals.config._atoms_cache.values())
        distances = {}
        for x in itertools.combinations(ids, 2):
            distances[f"{tuple(x)}"] = []
        for atoms in atoms_lst:
            positions = atoms.get_positions()
            for x in itertools.combinations(ids, 2):
                distances[f"{tuple(x)}"].append(
                    np.linalg.norm(positions[x[0]] - positions[x[1]])
                )

        df = pd.DataFrame({"step": list(range(len(atoms_lst)))} | distances)

        fig = px.line(
            df,
            x="step",
            y=df.columns,
            title="Distance between selected particles",
            render_mode="svg"  # This is important, otherwise openGL will be used
            # and there can/will be issues with three.js
        )
        if self.smooth:
            smooth_df = df.rolling(window=100).mean().dropna()
            for col in smooth_df.columns:
                if col != "step":
                    fig.add_scatter(
                        x=smooth_df["step"], y=smooth_df[col], name=f"smooth_{col}"
                    )
        return fig


class Properties2D(BaseModel):
    x_data: str = "step"
    y_data: str = "energy"
    color: str = "energy"
    fix_aspect_ratio: bool = True

    @classmethod
    def schema_from_atoms(cls, atoms):
        schema = cls.schema()
        available_properties = list(atoms[0].calc.results.keys())
        available_properties += ["step"]
        schema["properties"]["x_data"]["enum"] = available_properties
        schema["properties"]["y_data"]["enum"] = available_properties
        schema["properties"]["color"]["enum"] = available_properties
        return schema

    def run(self, ids):
        print(f"run {self}")
        atoms_lst = list(globals.config._atoms_cache.values())

        if self.x_data == "step":
            x_data = list(range(len(atoms_lst)))
        else:
            x_data = [x.calc.results[self.x_data] for x in atoms_lst]

        if self.y_data == "step":
            y_data = list(range(len(atoms_lst)))
        else:
            y_data = [x.calc.results[self.y_data] for x in atoms_lst]

        if self.color == "step":
            color = list(range(len(atoms_lst)))
        else:
            color = [x.calc.results[self.color] for x in atoms_lst]

        df = pd.DataFrame({self.x_data: x_data, self.y_data: y_data, self.color: color})
        fig = px.scatter(
            df, x=self.x_data, y=self.y_data, color=self.color, render_mode="svg"
        )
        if self.fix_aspect_ratio:
            fig.update_yaxes(
                scaleanchor="x",
                scaleratio=1,
            )

        return fig


class Properties1D(BaseModel):
    value: str = "energy"
    smooth: bool = False

    @classmethod
    def schema_from_atoms(cls, atoms):
        schema = cls.schema()
        available_properties = list(atoms[0].calc.results.keys())
        schema["properties"]["value"]["enum"] = available_properties
        return schema

    def run(self, ids):
        atoms_lst = list(globals.config._atoms_cache.values())

        data = np.array([x.calc.results[self.value] for x in atoms_lst])

        df = pd.DataFrame({"step": list(range(len(atoms_lst))), self.value: data})

        fig = px.line(df, x="step", y=self.value, render_mode="svg")

        if self.smooth:
            smooth_df = df.rolling(window=100).mean().dropna()
            for col in smooth_df.columns:
                if col != "step":
                    fig.add_scatter(
                        x=smooth_df["step"], y=smooth_df[col], name=f"smooth_{col}"
                    )

        return fig



from zndraw import globals


def serialize_atoms(start: int, stop: int):
    result = {"position": [], "force": [], "box": []}
    try:
        for step in range(start, stop):
            atoms = globals.config.get_atoms(step=int(step))
            result["position"].append(atoms.get_positions().tolist())
            result["box"].append(atoms.get_cell().diagonal().tolist())
            # TODO MAKE THIS OPTIONAL!!, also energy, etc.
            # try:
            #     result["force"].append(atoms.get_forces().tolist())
            # except:
            #     result["force"].append(np.zeros_like(atoms.get_positions()).tolist())
        return result
    except KeyError:
        return result



import dataclasses
import enum
import importlib
import pathlib
import typing

import ase.io
import pydantic
import tqdm
import znh5md
from pydantic import BaseModel, Field, PrivateAttr


class CameraChoices(str, enum.Enum):
    OrthographicCamera = "OrthographicCamera"
    PerspectiveCamera = "PerspectiveCamera"


class Config(BaseModel):
    file: str = Field(..., description="Trajectory File")
    camera: CameraChoices = Field(
        CameraChoices.PerspectiveCamera, description="Camera style"
    )
    bond_size: float = Field(1.0, description="Bond Size")
    sphere_size: float = Field(1.0, description="Sphere Size")
    resolution: int = Field(10, description="Resolution")
    max_fps: int = Field(100, description="Maximum Frames Per Second")
    frames_per_post: int = Field(100, description="Frames per JS POST request")
    active_update_function: str = Field(None, description="Active Update Function")
    material: str = Field("MeshPhongMaterial", description="Material")
    antialias: bool = Field(True, description="Antialias")
    continuous_loading: bool = Field(
        False, description="Continuous Loading of the trajectory"
    )

    _atoms_cache = PrivateAttr(default_factory=dict)
    _modifier_applied: bool = PrivateAttr(False)

    def get_modifier_schema(self, update_function) -> dict:
        """Add an update function to the config.

        Arguments
        ---------
        update_function: str
            The name of the update function 'module.cls'

        Returns
        -------
        dict: The signature of the function.
        """
        module_name, function_name = update_function.rsplit(".", 1)
        module = importlib.import_module(module_name)
        instance: pydantic.BaseModel = getattr(module, function_name)()
        schema = instance.schema()
        schema["title"] = update_function
        return schema

    def run_modifier(self, modifier, selected_ids, step, modifier_kwargs, **kwargs):
        """Run the update function.

        Arguments
        ---------
        update_function: str
            The name of the update function 'module.cls'
        selected_ids: list
            The selected ids.
        step: int
            The current step.

        Returns
        -------
        Save the updated atoms in the cache.
        """
        self._modifier_applied = True
        module_name, function_name = modifier.rsplit(".", 1)
        module = importlib.import_module(module_name)
        cls: pydantic.BaseModel = getattr(module, function_name)
        instance = cls(**modifier_kwargs)
        atoms = instance.run(selected_ids, self.get_atoms(step=step), **kwargs)
        for key in list(self._atoms_cache.keys()):
            # we remove all steps after the current one
            if key > step:
                del self._atoms_cache[key]

        for idx, atom in enumerate(atoms):
            self._atoms_cache[idx + step + 1] = atom

    def get_atoms(self, step: int = 0) -> ase.Atoms:
        """Get the atoms for a given step.

        Raises:
            KeyError: If the step could not be loaded.
        """
        try:
            return self._atoms_cache[step]
        except KeyError:
            self.load_atoms(step)
        return self._atoms_cache[step]

    def load_atoms(self, step: int = 999999999):
        """Load the atoms up to a given step."""
        # TODO ZnH5MD
        if self._modifier_applied:
            return  # We don't want to load any further from file at this point
        if pathlib.Path(self.file).suffix == ".h5":
            # We load all at once here
            self._atoms_cache.update(
                dict(enumerate(znh5md.ASEH5MD(self.file).get_atoms_list()[: step + 1]))
            )
        else:
            for idx, atoms in enumerate(tqdm.tqdm(ase.io.iread(self.file))):
                self._atoms_cache[idx] = atoms
                if idx == step:
                    break


config: Config = None

graph = None
atoms = None
